/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "Graph.h"

#include <Components/SubscriptionsPanel/SubscriptionsPanel.h>
#include <Core/MessageBroker/MessageBroker.h>

#include <QDebug>
#include <QTimer>

Graph::Graph() : Module(ModuleId::GRAPH)
{
    setupUi();
    customContextMenuActionSetup();

    connect(&updaterTimer, &QTimer::timeout, this, &Graph::onUpdateTimerTick);
    updaterTimer.setSingleShot(false);
    updaterTimer.start(updatePeriod);
}

XmlObject Graph::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("stopped", stopped ? 1 : 0);
    obj.addAttribute("following", following ? 1 : 0);
    obj.addAttribute("y_lower_range", (float)plot->yAxis->range().lower);
    obj.addAttribute("y_upper_range", (float)plot->yAxis->range().upper);
    obj.addAttribute("x_lower_range", (float)plot->xAxis->range().lower);
    obj.addAttribute("x_upper_range", (float)plot->xAxis->range().upper);

    for (auto filter : lines.keys())
    {
        XmlObject element("subscription");
        element.addAttribute("filter", filter.toString());
        obj.addChild(element);
    }

    return obj;
}

void Graph::fromXmlObject(const XmlObject& obj)
{
    int tmp;
    obj.getAttribute("stopped", tmp);
    stopped = (tmp == 1);
    obj.getAttribute("following", tmp);
    following = (tmp == 1);

    float lowerRange, upperRange;
    obj.getAttribute("y_lower_range", lowerRange);
    obj.getAttribute("y_upper_range", upperRange);
    plot->yAxis->setRange(lowerRange, upperRange);
    obj.getAttribute("x_lower_range", lowerRange);
    obj.getAttribute("x_upper_range", upperRange);
    plot->xAxis->setRange(lowerRange, upperRange);

    for (int i = 0; i < obj.childCount(); i++)
    {
        XmlObject child = obj.childAt(i);

        if (child.getObjectName() == "subscription")
        {
            auto filter = Filter::fromString(child.getAttribute("filter"));
            onFilterAdded(filter);
        }
    }
}

void Graph::onSubscribeClicked()
{
    SubscriptionsPanel* panel = new SubscriptionsPanel(lines.keys());
    panel->setWindowTitle("Graph subscriptions");
    connect(panel, &SubscriptionsPanel::filterAdded, this,
            &Graph::onFilterAdded);
    connect(panel, &SubscriptionsPanel::filterRemoved, this,
            &Graph::onFilterRemoved);
    panel->show();
}

void Graph::onClearClicked()
{
    for (auto filter : lines.keys())
    {
        for (auto field : lines[filter].first.keys())
            lines[filter].first[field].graph->data()->clear();
    }

    plot->replot();
}

void Graph::onStopClicked(QAction* action, bool checked)
{
    stopped = checked;
    action->setChecked(stopped);

    if (stopped)
        updaterTimer.stop();
    else
        updaterTimer.start(updatePeriod);
}

void Graph::onFollowClicked(bool checked) { following = checked; }

void Graph::onFilterAdded(const Filter& filter)
{
    // Fail if the filter is already in the list
    if (lines.keys().contains(filter))
        return;

    // Add a new line for each field of the filter
    for (auto field : filter.getFields())
    {
        if (field == "timestamp")
            continue;

        // Create the graph
        QCPGraph* graph = plot->addGraph();
        graph->setPen(
            QPen(QColor(rand() % 200, rand() % 200, rand() % 200), 2));
        graph->setScatterStyle(QCPScatterStyle(QCPScatterStyle::ssNone, 2));
        graph->setSelectable(QCP::stNone);
        graph->setName(field);

        // Add the filter and the graph along with new buffers to the lists
        lines[filter].first[field] = {graph, QVector<double>(),
                                      QVector<double>()};
    }
    plot->replot();

    MessageBroker::getInstance().subscribe(
        filter, this,
        [this](const Message& message, const Filter& filter)
        { onMessageReceived(message, filter); });
}

void Graph::onFilterRemoved(const Filter& filter)
{
    if (lines.contains(filter))
    {
        for (auto field : lines[filter].first.keys())
        {
            auto line = lines[filter].first[field];

            plot->removeGraph(line.graph);
            plot->replot();
        }

        lines.remove(filter);

        MessageBroker::getInstance().unsubscribe(filter, this);
    }
    else
    {
        qWarning() << "Removing filter not found in the graph";
    }
}

void Graph::onUpdateTimerTick()
{
    bool newData = false;
    double maxX  = 0;

    // Check if new data items are available and redraw
    for (auto filter : lines.keys())
    {
        // Check if the data have been updated
        if (lines[filter].second)
        {
            for (auto field : lines[filter].first.keys())
            {
                auto& line = lines[filter].first[field];

                if (line.bufferX.size() > 0)
                {
                    newData = true;

                    line.graph->addData(line.bufferX, line.bufferY);

                    for (auto tmp : line.bufferX)
                        if (tmp > maxX)
                            // cppcheck-suppress useStlAlgorithm
                            maxX = tmp;

                    line.bufferX.clear();
                    line.bufferY.clear();

                    // Check if there are too many data points
                    line.graph->data()->removeBefore(maxX - MAX_DATA_AGE);
                }
            }
        }
    }

    if (newData)
    {
        plot->replot();

        if (following)
            plot->xAxis->setRange(maxX, plot->xAxis->range().size(),
                                  Qt::AlignmentFlag::AlignRight);
    }
}

void Graph::onMessageReceived(const Message& message, const Filter& filter)
{
    if (stopped && !lines.contains(filter))
        return;

    if (!message.getFields().contains("timestamp"))
        return;

    for (auto field : filter.getFields())
    {
        if (field == "timestamp")
            continue;

        auto& line               = lines[filter].first[field];
        QVector<double>& bufferX = line.bufferX;
        QVector<double>& bufferY = line.bufferY;

        double x = message.getField("timestamp").getUnsignedInteger() / 1e6;
        double y = message.getField(field).getDouble();

        lines[filter].first[field].graph->setName(field + ": " +
                                                  QString::number(y));

        // Check if the timestamp resets
        if (lastTimestamp - 1 > x)
            onClearClicked();

        lastTimestamp = x;
        bufferX.append(x);
        bufferY.append(y);
    }

    // Flag the data as updated
    lines[filter].second = true;
}

void Graph::setupUi()
{
    QHBoxLayout* outerLayout = new QHBoxLayout;
    outerLayout->setContentsMargins(0, 0, 0, 0);

    plot = new QCustomPlot;
    outerLayout->addWidget(plot);

    setLayout(outerLayout);

    QFont pfont = QApplication::font();
    plot->legend->setIconSize(32, 2 * legendFontSize - 2);
    pfont.setPointSize(legendFontSize);
    plot->legend->setFont(pfont);

    plot->axisRect()->insetLayout()->setInsetAlignment(
        0, Qt::AlignTop | Qt::AlignLeft);
    plot->legend->setVisible(true);

    // Allow user to drag axis ranges with mouse, zoom with mouse wheel and
    // select graphs by clicking:
    plot->setInteractions(QCP::iRangeDrag | QCP::iRangeZoom |
                          QCP::iSelectPlottables);

    plot->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(plot, &QCustomPlot::customContextMenuRequested, this,
            &QWidget::customContextMenuRequested);
}

void Graph::customContextMenuActionSetup()
{
    QAction* subscriptions = new QAction("Manage subscriptions");
    subscriptions->setToolTip("Edit graph series");
    connect(subscriptions, &QAction::triggered, this,
            &Graph::onSubscribeClicked);
    customContextMenuActions.append(subscriptions);

    QAction* clear = new QAction("Clear");
    clear->setToolTip("Removes all data currently on the plot");
    connect(clear, &QAction::triggered, this, &Graph::onClearClicked);
    customContextMenuActions.append(clear);

    QAction* stop = new QAction("Stop");
    stop->setCheckable(true);
    stop->setChecked(stopped);
    stop->setToolTip("Pauses the graph");
    connect(stop, &QAction::triggered, this,
            [this, stop](bool checked) { onStopClicked(stop, checked); });
    customContextMenuActions.append(stop);

    QAction* follow = new QAction("Follow");
    follow->setCheckable(true);
    follow->setChecked(following);
    follow->setToolTip("Automatically move the graph to follow new data");
    connect(follow, &QAction::triggered, this, &Graph::onFollowClicked);
    customContextMenuActions.append(follow);
}
