/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "ValuesViewerConfigPanel.h"

#include <QColorDialog>
#include <QDebug>

#include "ui_ValuesViewerConfigPanel.h"

ValuesViewerConfigPanel::ValuesViewerConfigPanel(QWidget* parent)
    : QWidget(parent), ui(new Ui::ValuesViewerConfigPanel)
{
    ui->setupUi(this);
    connect(ui->add_pushButton, &QPushButton::clicked, this,
            &ValuesViewerConfigPanel::onAddFieldClicked);
    connect(ui->edit_pushButton, &QPushButton::clicked, this,
            &ValuesViewerConfigPanel::onEditClicked);
    connect(ui->remove_pushButton, &QPushButton::clicked, this,
            &ValuesViewerConfigPanel::onRemoveClicked);
    connect(ui->pickColor_pushButton, &QPushButton::clicked, this,
            &ValuesViewerConfigPanel::onPickColordClicked);
    connect(ui->save_pushButton, &QPushButton::clicked, this,
            &ValuesViewerConfigPanel::onSaveClicked);
    connect(ui->emit_radioButton, &QRadioButton::toggled, this,
            &ValuesViewerConfigPanel::onEmitRadioButtonToggled);
    setAttribute(Qt::WA_DeleteOnClose, true);
}

ValuesViewerConfigPanel::~ValuesViewerConfigPanel()
{
    // The QRadioButtons are destroyed automatically.
    delete ui;
}

void ValuesViewerConfigPanel::onAddFieldClicked()
{
    ValueElement e = createRuleFromControls();
    if (e.getName() != "")
    {
        int index = computeViewIndex(e.getName());
        addRule(index, e);
    }
}

ValueElement ValuesViewerConfigPanel::createRuleFromControls()
{
    QString name           = ui->name_lineEdit->text().trimmed();
    QString topic          = ui->topic_lineEdit->text().trimmed();
    QString receivedValue  = ui->receivedValue_lineEdit->text().trimmed();
    QString displayedValue = ui->displayedValue_lineEdit->text().trimmed();
    QString color          = ui->color_lineEdit->text().trimmed();
    QString outpuTopic     = ui->outputTopic_lineEdit->text().trimmed();
    bool isDisplayView     = ui->displayInView_checkBox->isChecked();
    ValueElement e(name, topic, receivedValue, displayedValue, color,
                   outpuTopic);
    e.setDisplayInView(isDisplayView);
    return e;
}

void ValuesViewerConfigPanel::onEditClicked()
{
    ValueElement e = createRuleFromControls();
    if (e.getName() != "")
    {
        int index = findSelectedView();
        if (index >= 0 && index < rules.count())
        {
            rules[index] = e;
            views[index]->setText(e.toString());
        }
    }
}

void ValuesViewerConfigPanel::onRemoveClicked()
{
    int selectedView = findSelectedView();
    if (selectedView >= 0 && selectedView < rules.count())
    {
        rules.removeAt(selectedView);
    }
    if (selectedView >= 0 && selectedView < views.count())
    {
        delete views[selectedView];
        views.removeAt(selectedView);
    }
}

void ValuesViewerConfigPanel::onPickColordClicked()
{
    QColor initial(ui->color_lineEdit->text().trimmed());
    QColor color = QColorDialog::getColor(initial, this);
    ui->color_lineEdit->setText(color.name());
}

void ValuesViewerConfigPanel::onValueElementClicked()
{
    int selectedView = findSelectedView();
    if (selectedView >= 0 && selectedView < rules.count())
    {
        setTextBoxValues(rules[selectedView]);
    }
}

void ValuesViewerConfigPanel::onSaveClicked()
{
    emit configurationSaved(this);
    this->close();
}

int ValuesViewerConfigPanel::computeViewIndex(const QString& viewName)
{
    int i = 0;
    for (; i < rules.count(); i++)
    {

        if (rules.at(i).getName().compare(viewName) >= 0)
        {
            return i;
        }
    }
    return i;
}

QRadioButton* ValuesViewerConfigPanel::createView(const ValueElement& el)
{
    return new QRadioButton(el.toString());
}

void ValuesViewerConfigPanel::setTextBoxValues(const ValueElement& el)
{
    ui->name_lineEdit->setText(el.getName());
    ui->topic_lineEdit->setText(el.getTopic());
    ui->receivedValue_lineEdit->setText(el.getReceivedValue());
    ui->displayedValue_lineEdit->setText(el.getDisplayedValue());
    ui->color_lineEdit->setText(el.getColor());

    ui->displayInView_checkBox->setChecked(el.isDisplayInView());

    if (el.isEmitActive())
    {
        ui->outputTopic_lineEdit->setText(el.getOutputTopic());
        ui->emit_radioButton->setChecked(true);
    }
    else
    {
        ui->doNotEmit_radioButton->setChecked(true);
        ui->outputTopic_lineEdit->setText("");
    }
}

int ValuesViewerConfigPanel::findSelectedView() const
{
    for (int i = 0; i < views.count(); i++)
    {
        if (views[i]->isChecked())
        {
            return i;
        }
    }
    return -1;
}

void ValuesViewerConfigPanel::addRule(int index, const ValueElement& el)
{
    rules.insert(index, el);
    QRadioButton* view = createView(el);
    views.insert(index, view);
    ui->fieldsLayout->insertWidget(index, view);
    connect(view, &QRadioButton::clicked, this,
            &ValuesViewerConfigPanel::onValueElementClicked);
}

void ValuesViewerConfigPanel::onEmitRadioButtonToggled(bool val)
{
    ui->outputTopic_lineEdit->setEnabled(val);

    if (val == false)
    {
        ui->outputTopic_lineEdit->setText("");
    }
}

void ValuesViewerConfigPanel::setRules(const QList<ValueElement>& newValues)
{
    for (int i = 0; i < newValues.count(); i++)
    {
        addRule(i, newValues[i]);
    }
}

QList<ValueElement> ValuesViewerConfigPanel::getRules() const { return rules; }

int ValuesViewerConfigPanel::getColumnsCount() const
{
    return ui->columns_spinBox->value();
}

void ValuesViewerConfigPanel::setColumnsCount(int value)
{
    return ui->columns_spinBox->setValue(value);
}
